"""This module contains all the algorithms and formulas necesary for the 
calculation of the sun-path charts. All the formulas, algorithms and procedures
are based on the paper:
'Solar position algorithm for solar radiation applications'
(I. Reda, A. Andreas, 2003)"""
from datetime import datetime, timedelta
import math
from math import degrees as deg
from math import radians as rad
import sun_path_diagrams.tables as tb


class SunPosition:
    """Formulas and algorithms to calculate the sun position. The methods, the 
    variable names and the sequence of calculations follows (mostly) the one 
    found in the paper 
    'Solar position algorithm for solar radiation applications' 
    (I. Reda, A. Andreas, 2003)."""
    def __init__(self, lat, lon, date_time, timezone, obs_elev, press, temp):
        self.lat = lat  # observer latitude
        self.lon = lon  # observer longitude
        self.date_time = date_time  # local date and time
        self.timezone = timezone  # time zone
        self.obs_elev = obs_elev  # observer elevation in meters
        self.press = press  # atmospheric pressure in mbar
        self.temp = temp  # temperature in Celcius degrees

    def sun_position(self):
        """
        Returns a dictionary holding the azimuth and altitude angles for
        every day in the given year. Keys in the dictionary follows the format
        0203 (Feb the 3rd).
        :returns: a dictionary with 365/366 items
        """
        dt_obj = datetime.fromisoformat(self.date_time)
        timezone = self.timezone

        # convert local date and time to universal time (UT)
        # note ut_day carry a decimal to acount for the hours and minutes
        ut_year, ut_month, ut_day = self.__lct2ut(dt_obj, timezone)
        positions = {}

        firstday = datetime(ut_year, 1, 1)
        lastday = datetime(ut_year, 12, 31)
        # create a list holding 365/366 days for the given year
        alldays = [firstday + timedelta(days=x) for x in
                   range((lastday - firstday).days + 1)]

        for dt in alldays:
            year = dt.year
            month = dt.month
            day = dt.day

            azim_lst = []
            alt_lst = []

            if month == ut_month and day == int(ut_day):
                key = dt.strftime('p%b%d')
            else:
                key = dt.strftime('%b%d')

            for i in range(0, 25):  # 25 hours
                decimal_day = day + i/24  # 24 hours

                elev_azim = self.__elev_azim(year, month, decimal_day)

                azim_lst.append(elev_azim['topo_azim_ang'])
                alt_lst.append(elev_azim['topo_elev_ang'])

            positions[key] = (azim_lst, alt_lst)

        return positions

    def sun_position_point(self):
        """
        Returns a tuple holding the azimuth and altitude angles for the user 
        selected date and time.
        :returns: a tuple holding azimuth and altitude
        """
        dt_obj = datetime.fromisoformat(self.date_time)
        timezone = self.timezone

        # convert local date and time to universal time (UT)
        # note ut_day carry a decimal to acount for the hours and minutes
        ut_year, ut_month, ut_day = self.__lct2ut(dt_obj, timezone)
        elev_azim = self.__elev_azim(ut_year, ut_month, ut_day)
        azim = elev_azim['topo_azim_ang']
        alt = elev_azim['topo_elev_ang']

        return (azim, alt)

    def __elev_azim(self, year, month, day):
        """
        Gets the azimuth and elevation (also known as altitude)
        :param year: year - universal time
        :param month: month - universal time
        :param day: day - universal time
        :returns: a dictionary holding the sun's azimuth and the elevation
        """
        # calculate the Julian dates
        jdates = self.__juliandates(year, month, day)
        jd = jdates['julianday']
        jc = jdates['juliancentury']
        jce = jdates['julian_ephemeris_century']
        jme = jdates['julian_ephemeris_millenium']

        # calculate the heliocentric parameters
        hc = self.__heliocentric_params(jme)
        L = hc['heliocentric_lon']
        B = hc['heliocentric_lat']
        R = hc['earth_radius']

        # calculate the geocentric latitude and longitude
        glatlon = self.__geocentric_latlon(B, L)
        geocentric_lat = glatlon['geocentric_lat']
        geocentric_lon = glatlon['geocentric_lon']

        # calculate the nutation in longitude and obliquity
        nto = self.__nutations(jce)
        nutation_longitude = nto['nutation_longitude']
        nutation_obliquity = nto['nutation_obliquity']

        # calculate the true obliquity of the ecliptic
        true_obliquity = self.__true_obliq_ecliptic(jme, nutation_obliquity)

        # calculate the apparent sun longitude
        app_sun_longitude = self.__apparent_app_sun_lon(geocentric_lon,
                                                        nutation_longitude, R)

        # calculate the apparent sidereal time at Greenwich
        app_sidereal_time = self.__apparent_sidereal_time(jd, jc,
                                                          nutation_longitude,
                                                          true_obliquity)

        # calculate the geocentric right ascension
        geocentric_right_asc = self.__geocentric_right_asc(app_sun_longitude,
                                                           true_obliquity,
                                                           geocentric_lat)

        # calculate the geocentric sun declination
        geocentric_declination = self.__geocentric_declination(
                                                             geocentric_lat,
                                                             true_obliquity,
                                                             app_sun_longitude)

        # calculate the observer local hour angle
        local_hour_angle = self.__local_hour_angle(app_sidereal_time, self.lon,
                                                   geocentric_right_asc)

        # calculate the topocentric declination and
        # topocentric local hour angle
        ta = self.__topocentric_angles(R, self.lat, self.obs_elev,
                                       local_hour_angle,
                                       geocentric_declination,
                                       geocentric_right_asc)
        topo_declination = ta['topo_declination']
        topo_localhour_ang = ta['topo_localhour_ang']

        # calculate the topocentric elevation and azimuth angles
        return self.__elevation_azimuth(self.lat, topo_declination,
                                        topo_localhour_ang, self.press,
                                        self.temp)

    def __juliandates(self, year, month, day):
        """
        Calculates the Julian day (jd), Julian Century (jc),
        Julian Ephemeris Century (jce) and Julian Ephemeris Millenium (jme)
        :param year: year - universal time
        :param month: month - universal time
        :param day: day - universal time
        :returns: a dictionary holding all the Julian dates
        """
        # delta_t is the difference between earth rotation time and terrestrial 
        # time. How it is calculated here differs from the method recommended 
        # in the paper
        delta_t = self.__delta_t(year, month)

        if month < 3:
            year = year - 1
            month = month + 12

        # a and b are intermediate calculation variables
        a = int(year/100)
        b = 2 - a + int(a/4)

        julianday = (int(365.25*(year + 4716)) + int(30.6001*(month + 1)) +
                     day + b - 1524.5)

        julian_ephday = julianday + delta_t/86400  # julian ephemeris day
        juliancentury = (julianday - 2451545)/36525
        julian_ephcentury = (julian_ephday - 2451545)/36525
        julian_ephmillenium = julian_ephcentury/10

        return {'julianday': julianday,
                'juliancentury': juliancentury,
                'julian_ephemeris_century': julian_ephcentury,
                'julian_ephemeris_millenium': julian_ephmillenium}

    def __heliocentric_params(self, jme):
        """
        Gets the earth's heliocentric longitude, latitude and radius vector
        :param jme: julian_ephemeris_millenium
        :returns: a dictionary holding the heliocentric parameters
        """
        # earth heliocentric longitude
        l0 = self.__calc_l0(tb.l0_A, tb.l0_B, tb.l0_C, jme)
        l1 = self.__calc_l0(tb.l1_A, tb.l1_B, tb.l1_C, jme)
        l2 = self.__calc_l0(tb.l2_A, tb.l2_B, tb.l2_C, jme)
        l3 = self.__calc_l0(tb.l3_A, tb.l3_B, tb.l3_C, jme)
        l4 = self.__calc_l0(tb.l4_A, tb.l4_B, tb.l4_C, jme)
        l5 = self.__calc_l0(tb.l5_A, tb.l5_B, tb.l5_C, jme)

        l_rad = (l0 + l1*jme + l2*pow(jme, 2) + l3*pow(jme, 3) +
                 l4*pow(jme, 4) + l5*pow(jme, 5))/pow(10, 8)

        helio_lon = self.__limit_360(deg(l_rad))

        # earth heliocentric latitude
        b0 = self.__calc_l0(tb.b0_A, tb.b0_B, tb.b0_C, jme)
        b1 = self.__calc_l0(tb.b1_A, tb.b1_B, tb.b1_C, jme)

        b_rad = (b0 + b1*jme)/pow(10, 8)
        helio_lat = deg(b_rad)

        # earth radius vector
        r0 = self.__calc_l0(tb.r0_A, tb.r0_B, tb.r0_C, jme)
        r1 = self.__calc_l0(tb.r1_A, tb.r1_B, tb.r1_C, jme)
        r2 = self.__calc_l0(tb.r2_A, tb.r2_B, tb.r2_C, jme)
        r3 = self.__calc_l0(tb.r3_A, tb.r3_B, tb.r3_C, jme)
        r4 = self.__calc_l0(tb.r4_A, tb.r4_B, tb.r4_C, jme)

        earth_radius = (r0 + r1*jme + r2*pow(jme, 2) + r3*pow(jme, 3) +
                        r4*pow(jme, 4))/pow(10, 8)

        return {'heliocentric_lon': helio_lon,
                'heliocentric_lat': helio_lat,
                'earth_radius': earth_radius}

    def __geocentric_latlon(self, helio_lat, helio_lon):
        """
        Calculates geocentric latitude and longitude,
        :param helio_lat: heliocentric latitude
        :param helio_lon: heliocentric longitude
        :returns: a dictionary holding the geocentric latitude and longitude
        """
        geo_lat = -helio_lat
        geo_lon = self.__limit_360(helio_lon + 180)

        return {'geocentric_lon': geo_lon,
                'geocentric_lat': geo_lat}

    def __nutations(self, jce):
        """
        Calculates the nutation in longitude and obliquity
        :param jce: Julian Ephemeris Century
        :returns: a dictionary holding the nutation in longitude and obliquity
        """
        x0 = (297.85036 + 445267.111480*jce - 0.0019142*pow(jce, 2) +
              pow(jce, 3)/189474)
        x1 = (357.52772 + 35999.050340*jce - 0.0001603*pow(jce, 2) -
              pow(jce, 3)/300000)
        x2 = (134.96298 + 477198.867398*jce + 0.0086972*pow(jce, 2) +
              pow(jce, 3)/56250)
        x3 = (93.27191 + 483202.017538*jce - 0.0036825*pow(jce, 2) +
              pow(jce, 3)/327270)
        x4 = (125.04452 - 1934.136261*jce + 0.0020708*pow(jce, 2) +
              pow(jce, 3)/450000)

        table2 = zip(tb.y0, tb.y1, tb.y2, tb.y3, tb.y4, tb.a, tb.b, tb.c, tb.d)
        tmp_20 = 0  # formula #20
        tmp_21 = 0  # formula #21

        for i in table2:
            xy = x0*i[0] + x1*i[1] + x2*i[2] + x3*i[3] + x4*i[4]
            tmp_20 = tmp_20 + (i[5] + i[6]*jce)*math.sin(xy)
            tmp_21 = tmp_21 + (i[7] + i[8]*jce)*math.cos(xy)

        nut_lon = tmp_20/36000000
        nut_obl = tmp_21/36000000

        return {'nutation_longitude': nut_lon,
                'nutation_obliquity': nut_obl}

    def __true_obliq_ecliptic(self, jme, nut_obl):
        """
        Calculates the true obliquity of the ecliptic
        :param jme: Julian Ephemeris Millenium
        :param nut_obl: nutation in longitude
        :returns: the true obliquity of the ecliptic in degrees
        """
        u = jme/10
        e0 = (84381.448 - 4680.93*u - 1.55*pow(u, 2) + 1999.25*pow(u, 3) -
              51.38*pow(u, 4) - 249.67*pow(u, 5) - 39.05*pow(u, 6) +
              7.12*pow(u, 7) + 27.87*pow(u, 8) + 5.79*pow(u, 9) +
              2.45*pow(u, 10))

        return e0/3600 + nut_obl

    def __apparent_app_sun_lon(self, geo_lon, nut_lon, earth_radius):
        """
        Calculates the apparent sun longitude
        :param geo_lon: geocentric longitude
        :param nut_lon: nutation in longitude
        :param earth_radius: earth_radius
        :returns: apparent sun longitude in degrees
        """
        return geo_lon + nut_lon - 20.4898/(3600*earth_radius)

    def __apparent_sidereal_time(self, jd, jc, nut_lon, true_obliq_eclip):
        """
        Calculates the apparent sidereal time at Greenwich
        :param jd: julian day
        :param jc: julian century
        :param nut_lon: nutation in longitude
        :param true_obliq_eclip: true obliquity of the ecliptic
        :returns: apparent sidereal time in degrees
        """
        v0_ = (280.46061837 + 360.98564736629*(jd - 2451545) +
               0.000387933*pow(jc, 2) - pow(jc, 3)/38710000)
        v0 = self.__limit_360(v0_)  # v0 is the mean sidereal time at Greenwich

        return v0 + nut_lon*math.cos(rad(true_obliq_eclip))

    def __geocentric_right_asc(self, app_sun_lon, true_obliq_eclip, geo_lat):
        """
        Calculates the geocentric sun right ascension
        :param app_sun_lon: apparent sun longitude
        :param true_obliq_eclip: true obliquity of the ecliptic
        :param geo_lat: geocentric latitude
        :returns: sun right ascension in degrees
        """
        app_sun_lon = rad(app_sun_lon)
        true_obliq_eclip = rad(true_obliq_eclip)
        geo_lat = rad(geo_lat)

        a = math.sin(app_sun_lon)*math.cos(true_obliq_eclip)
        b = math.tan(geo_lat)*math.sin(true_obliq_eclip)
        y = a - b
        x = math.cos(app_sun_lon)
        tmp = deg(math.atan2(y, x))

        return self.__limit_360(tmp)

    def __geocentric_declination(self, geo_lat, true_obliq_eclip, app_sun_lon):
        """
        Calculates the geocentric sun declination
        :param geo_lat: geocentric latitude
        :param true_obliq_eclip: true obliquity of the ecliptic
        :param app_sun_lon: apparent sun longitude
        :returns: sun declination in degrees
        """
        geo_lat = rad(geo_lat)
        true_obliq_eclip = rad(true_obliq_eclip)
        app_sun_lon = rad(app_sun_lon)

        a = math.sin(geo_lat)*math.cos(true_obliq_eclip)
        b = math.cos(geo_lat)*math.sin(true_obliq_eclip)*math.sin(app_sun_lon)

        return deg(math.asin(a + b))

    def __local_hour_angle(self, sidereal_time, lon, geocentric_right_asc):
        """
        Calculates the observer local hour angle H
        :param sidereal_time: aparent sidereal time at Greewich
        :param lon: observer's longitude
        :param geocentric_right_asc: sun's geocentric right ascension
        :returns: local hour angle (H) in degrees
        """
        tmp = sidereal_time + lon - geocentric_right_asc

        return self.__limit_360(tmp)

    def __topocentric_angles(self, earth_radius, lat, obs_elev,
                             local_hour_angle, geocentric_declination,
                             geocentric_right_asc):
        """
        Calculates the topocentric sun right ascension,
        the topocentric sun declination, topocentric local hour angle,
        the topocentric zenith angle
        :param earth_radius: earth radius R
        :param lat: observer's latitude
        :param obs_elev: observer's elevation in meters
        :param local_hour_angle: observer's local hour angle
        :param geocentric_declination: geocentric sun declination
        :param geocentric_right_asc: geocentric sun's right ascension
        :returns: a dictionary holding the topocentric sun right ascension,
        the topocentric sun declination and topocentric local hour angle,
        all in degrees
        """
        lat = rad(lat)

        eq_hz_parall = rad(8.794/(3600*earth_radius))  # equatorial hz parallax
        u = math.atan(0.99664719*math.tan(lat))
        x = math.cos(u) + obs_elev/6378140*math.cos(lat)
        y = (0.99664719*math.sin(u) +
             obs_elev/6378140*math.sin(lat))

        # topocentric sun right ascension
        a = -x*math.sin(eq_hz_parall)*math.sin(rad(local_hour_angle))
        b = (math.cos(rad(geocentric_declination)) -
             x*math.sin(eq_hz_parall)*math.cos(rad(local_hour_angle)))
        parall_right_asc = deg(math.atan2(a, b))
        topo_right_asc = geocentric_right_asc + parall_right_asc

        # topocentric sun declination
        a1 = ((math.sin(rad(geocentric_declination)) -
              y*math.sin(eq_hz_parall))*math.cos(rad(parall_right_asc)))
        b1 = (math.cos(rad(geocentric_declination)) -
              y*math.sin(eq_hz_parall)*math.cos(rad(local_hour_angle)))
        topo_declination = deg(math.atan2(a1, b1))

        # topocentric local hour angle (H')
        topo_localhour_ang = local_hour_angle - parall_right_asc

        return {'topo_right_asc': topo_right_asc,
                'topo_declination': topo_declination,
                'topo_localhour_ang': topo_localhour_ang}

    def __elevation_azimuth(self, lat, topo_declination, topo_localhour_ang,
                            press, temp):
        """
        Calculates the topocentric elevation and azimuth angles
        :param earth_radius: earth radius R
        :param lat: observer's latitude
        :param topo_declination: topocentric sun declination
        :param topo_localhour_ang: topocentric local hour angle
        :param press: annual average local pressure in mbars
        :param temp: annual average local temeperature in Celcius degrees
        :returns: the topocentric sun right ascension
        """
        lat = rad(lat)
        topo_localhour_ang = rad(topo_localhour_ang)
        topo_declination = rad(topo_declination)

        # topocentric elevation angle
        a2 = math.sin(lat)*math.sin(topo_declination)
        b2 = math.cos(lat)*math.cos(topo_declination)*math.cos(topo_localhour_ang)
        e0 = deg(math.asin(a2 + b2))

        c2 = rad(e0 + 10.3/(e0 + 5.11))
        atm_corr = (press/1010)*(283/(273 + temp))*(1.02/(60*math.tan(c2)))
        topo_elev_ang = e0 + atm_corr  # topocentric elevation angle
        topo_zenith_ang = 90 - topo_elev_ang  # topocentric zenith angle

        # topocentric azimuth angle
        a3 = math.sin(topo_localhour_ang)
        b3 = (math.cos(topo_localhour_ang)*math.sin(lat) -
              math.tan(topo_declination)*math.cos(lat))
        c3 = deg(math.atan2(a3, b3))
        d3 = self.__limit_360(c3) + 180
        topo_azim_ang = self.__limit_360(d3)

        return {'topo_elev_ang': topo_elev_ang,
                'topo_azim_ang': topo_azim_ang}

    def __calc_l0(self, a, b, c, jme):
        """
        Calculates the formula # 10 in the paper
        :param a: A
        :param b: B
        :param c: C
        :param jme: julian_ephemeris_millenium
        :returns: term L0 (formula # 10)
        """
        if isinstance(a, tuple):
            tmp = zip(a, b, c)
            tmp_ = 0
            for i in tmp:
                tmp_ = tmp_ + i[0] * math.cos(i[1] + i[2]*jme)
        else:
            tmp_ = a * math.cos(b + c*jme)

        return tmp_

    def __limit_360(self, angle):
        """
        Gets a new angle in the range 0 to 360 degrees
        :param angle: angle to be re-calculated in degrees
        :returns: a new angle in the range 0 to 360
        """
        frac, _ = math.modf(angle/360)

        if angle > 0:
            out = 360 * abs(frac)
        else:
            out = 360 - 360 * abs(frac)

        return out

    def __juliandate(self, year, month, day):
        """Calculates the julian date for a Greenwich calendar date"""
        if month < 3:
            year = year - 1
            month = month + 12

        # Here we asume all years are > 1582
        a = int(year/100)
        b = 2 - a + int(a/4)

        if year < 0:
            c = int((365.25*year) - 0.75)
        else:
            c = int(365.25*year)

        d = int(30.6001*(month + 1))

        return b + c + d + day + 1720994.5

    def __juliandate2gcd(self, juliandate):
        """Converts julianday to Greenwich Calendar Date (gcd)"""
        jd = juliandate + 0.5
        frac, inte = math.modf(jd)

        if inte > 2299160:
            a = int((inte - 1867216.25)/36524.25)
            b = inte + a - int(a/4) + 1
        else:
            b = inte

        c = b + 1524
        d = int((c - 122.1)/365.25)
        e = int(365.25*d)
        g = int((c - e)/30.6001)
        day = c - e + frac - int(30.6001*g)

        if g < 13.5:
            month = g - 1
        else:
            month = g - 13

        if month > 2.5:
            year = d - 4716
        else:
            year = d - 4715

        return year, month, day

    def __lct2ut(self, dt_obj, timezone):
        """Converts local civil time (lct) to universal time (ut)"""
        year = dt_obj.year
        month = dt_obj.month
        day = dt_obj.day
        hour = dt_obj.hour
        minute = dt_obj.minute

        lct = hour + minute/60  # seconds are zero

        ut = lct - timezone
        day = day + ut/24
        juliandate = self.__juliandate(year, month, day)
        year_, month_, day_ = self.__juliandate2gcd(juliandate)

        return year_, month_, day_

    def __delta_t(self, year, month):
        """
        Calculates the delta T value. Formulas taken from:
        POLYNOMIAL EXPRESSIONS FOR DELTA T (ΔT)
        https://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html
        Note that only the formulas for year > 1600 are implemented.
        :param year: year
        :param month: month
        :returns: delta T in seconds
        """
        y = year + (month - 0.5/12)

        if year >= 1600 and year < 1700:
            t = y - 1600
            dt = 120 - 0.9808*t - 0.01532*pow(t, 2) + pow(t, 3)/7129
        elif year >= 1700 and year < 1800:
            t = y - 1700
            dt = (8.83 + 0.1603*t - 0.0059285*pow(t, 2) +
                  0.00013336*pow(t, 3) - pow(t, 4)/1174000)
        elif year >= 1800 and year < 1860:
            t = y - 1800
            dt = (13.72 - 0.332447*t + 0.0068612*pow(t, 2) +
                  0.0041116*pow(t, 3) - 0.00037436 * pow(t, 4) +
                  0.0000121272*pow(t, 5) - 0.0000001699*pow(t, 6) +
                  0.000000000875*pow(t, 7))
        elif year >= 1860 and year < 1900:
            t = y - 1860
            dt = (7.62 + 0.5737*t - 0.251754*pow(t, 2) + 0.01680668*pow(t, 3) -
                  0.0004473624*pow(t, 4) + pow(t, 5)/233174)
        elif year >= 1900 and year < 1920:
            t = y - 1900
            dt = (-2.79 + 1.494119*t - 0.0598939*pow(t, 2) +
                  0.0061966*pow(t, 3) - 0.000197*pow(t, 4))
        elif year >= 1920 and year < 1941:
            t = y - 1920
            dt = (21.20 + 0.84493*t - 0.076100*pow(t, 2) + 0.0020936*pow(t, 3))
        elif year >= 1941 and year < 1961:
            t = y - 1950
            dt = 29.07 + 0.407*t - pow(t, 2)/233 + pow(t, 3)/2547
        elif year >= 1961 and year < 1986:
            t = y - 1975
            dt = 45.45 + 1.067*t - pow(t, 2)/260 - pow(t, 3)/718
        elif year >= 1986 and year < 2005:
            t = y - 2000
            dt = (63.86 + 0.3345*t - 0.060374*pow(t, 2) + 0.0017275*pow(t, 3) +
                  0.000651814*pow(t, 4) + 0.00002373599*pow(t, 5))
        elif year >= 2005 and year < 2050:
            t = y - 2000
            dt = 62.92 + 0.32217*t + 0.005589*pow(t, 2)
        else:
            print('Error: year not in the range 1600 to 2050')

        return dt


if __name__ == '__main__':
    app = SunPosition(60, 0, '2003-01-01T00:00:00', 0, False, 0, 0, 0)
    pos = app.sun_position()


"""This module holds configuration variables for the charts"""

# figure size in inches
fig_zize = {'width': 8, 'height': 8}

# line symbol and labels for equinoxes and solstices
equisols = {'Jun21': ((0, (7, 1, 1, 1, 1, 1, 1, 1)), 'Jun 21'),
            'Mar21': ((0, (7, 1, 1, 1, 1, 1)), 'Mar-Sep 21'),
            'Dec21': ((0, (7, 1, 1, 1)), 'Dec 21')}

# equinoxes and solstices lines symbology properties
equisols_symb = {'color': 'black', 'linewidth': 1}

# user selected date and time
user_symb = {'color': '#eb9800', 'linewidth': 1, 'marker': '.'}

# user selected date and time (point symbol)
user_symb_point = {'color': 'red', 'marker': 'o', 'markersize': '6',
                   'linewidth': 0}

# analemmas symbology properties
ana_symb = {'color': 'black', 'linestyle': 'solid', 'linewidth': 0.5}

# hour symbology properties
hour_symb = {'color': 'black', 'size': '7', 'xytext': (-3, 5)}

# grid symbology properties
grid_symb = {'color1': '#cecece', 'color2': 'black', 'color3': '#767676',
             'linestyle': '-', 'linewidth': 0.5, 'angle': 52, 'size': '8'}


"This is the entry point of the package"
import sys
from sun_path_diagrams.gui import SunPathGUI


def main(args=None):
    """This is the entry point for the package"""
    if args is None:
        app = SunPathGUI()
        app.main_window.mainloop()


if __name__ == "__main__":
    sys.exit(main())


"""This module contains the Tkinter GUI code"""
import os
import csv
import tkinter as tk
import datetime
from tkinter import ttk
from tkinter import filedialog
from tkinter import messagebox
from sun_path_diagrams.algorithms_spa import SunPosition
from sun_path_diagrams.charting import PlotSunPath


class SunPathGUI():
    "GUI dialog for the input of parameters and running of the program"
    def __init__(self):
        self.main_window = tk.Tk()
        self.main_window.title('Sun Path Diagrams')
        self.main_window.resizable(False, False)
        self.main_window.configure(bg='#dcdad5')

        self.style = ttk.Style()
        self.style.theme_use('clam')

        self.__create_variables()
        self.__create_frames()
        self.__create_lbl_frames()
        self.__create_widgets()
        self.__create_layouts()

    def __create_variables(self):
        """Create StringVar variables"""
        dt = datetime.datetime.now()
        self.year = tk.StringVar(value=dt.year)
        self.month = tk.StringVar(value=dt.month)
        self.day = tk.StringVar(value=dt.day)
        self.hour = tk.StringVar(value=dt.hour)
        self.minute = tk.StringVar(value=dt.minute)
        self.timezone = tk.StringVar()
        self.daylight = tk.StringVar(value=0)
        self.latitude = tk.StringVar(value='51.05')
        self.longitude = tk.StringVar(value='-114.08')
        self.obs_elev = tk.StringVar(value='')
        self.press = tk.StringVar(value='')
        self.temp = tk.StringVar(value='')
        self.title = tk.StringVar(value='Chart title')
        self.hz_chart = tk.StringVar(value=1)
        self.vt_chart = tk.StringVar(value=1)
        self.plt_point = tk.StringVar(value=0)
        self.csv_data = tk.StringVar(value=0)
        self.outputdir = tk.StringVar(value=os.path.expanduser('~'))

    def __create_widgets(self):
        """Create widgets"""
        self.val = GUIValidation(self.main_window)
        # year
        self.year_lbl = ttk.Label(self.date_lbl_frame, text='Year: ')
        self.year_spin = ttk.Spinbox(self.date_lbl_frame, from_=1900, to=2100,
                                     increment=1, textvariable=self.year,
                                     width=5, validate='focusout',
                                     validatecommand=self.val.year_vcmd,
                                     invalidcommand=self.val.year_ivcmd)

        # month
        self.month_lbl = ttk.Label(self.date_lbl_frame, text='Month: ')
        self.month_spin = ttk.Spinbox(self.date_lbl_frame, from_=1, to=12,
                                      increment=1, textvariable=self.month,
                                      width=3, validate='focusout',
                                      validatecommand=self.val.month_vcmd,
                                      invalidcommand=self.val.month_ivcmd)

        # day
        self.day_lbl = ttk.Label(self.date_lbl_frame, text='Day: ')
        self.day_spin = ttk.Spinbox(self.date_lbl_frame, from_=1, to=31,
                                    increment=1, textvariable=self.day,
                                    width=3, validate='focusout',
                                    validatecommand=self.val.day_vcmd,
                                    invalidcommand=self.val.day_ivcmd)

        # hour
        self.hour_lbl = ttk.Label(self.date_lbl_frame, text='Hour: ')
        self.hour_spin = ttk.Spinbox(self.date_lbl_frame, from_=0, to=23,
                                     increment=1, textvariable=self.hour,
                                     width=3, validate='focusout',
                                     validatecommand=self.val.hour_vcmd,
                                     invalidcommand=self.val.hour_ivcmd)

        # minute
        self.minute_lbl = ttk.Label(self.date_lbl_frame, text='Minute: ')
        self.minute_spin = ttk.Spinbox(self.date_lbl_frame, from_=0, to=59,
                                       increment=1, textvariable=self.minute,
                                       width=3, validate='focusout',
                                       validatecommand=self.val.minute_vcmd,
                                       invalidcommand=self.val.minute_ivcmd)

        # time zone
        self.timezone_lbl = ttk.Label(self.date_lbl_frame, text='UTC: ')
        ut = ('-12', '-11', '-10', '-9', '-8', '-7', '-6', '-5', '-4', '-3',
              '-2', '-1', '0', '+1', '+2', '+3', '+4', '+5', '+6', '+7', '+8',
              '+9', '+10', '+11', '+12')
        self.timezone_dropdown = ttk.Combobox(self.date_lbl_frame,
                                              textvariable=self.timezone,
                                              state='readonly', justify='left',
                                              width=5, values=ut)
        self.timezone_dropdown.current(5)

        # daylight
        self.daylight_chkbox = ttk.Checkbutton(self.date_lbl_frame,
                                               text='Daylight Saving Time',
                                               variable=self.daylight,
                                               onvalue=1,
                                               offvalue=0)

        # latitud
        self.latitude_lbl = ttk.Label(self.location_lbl_frame,
                                      text='Latitude: ')
        self.latitude_entry = ttk.Entry(self.location_lbl_frame,
                                        textvariable=self.latitude, width=10,
                                        validate='focusout',
                                        validatecommand=self.val.lat_vcmd,
                                        invalidcommand=self.val.lat_ivcmd)

        # longitude
        self.longitude_lbl = ttk.Label(self.location_lbl_frame,
                                       text='Longitude: ')
        self.longitude_entry = ttk.Entry(self.location_lbl_frame,
                                         textvariable=self.longitude, width=10,
                                         validate='focusout',
                                         validatecommand=self.val.lon_vcmd,
                                         invalidcommand=self.val.lon_ivcmd)

        # observer elevation
        self.obs_elev_lbl = ttk.Label(self.other_lbl_frame,
                                      text='Observer elevation (meters): ')
        self.obs_elev_entry = ttk.Entry(self.other_lbl_frame,
                                        textvariable=self.obs_elev, width=11,
                                        validate='focusout',
                                        validatecommand=self.val.obs_elev_vcmd,
                                        invalidcommand=self.val.obs_elev_ivcmd)

        # pressure
        self.press_lbl = ttk.Label(self.other_lbl_frame,
                                   text='Annual avg local pressure (mbars): ')
        self.press_entry = ttk.Entry(self.other_lbl_frame,
                                     textvariable=self.press, width=11,
                                     validate='focusout',
                                     validatecommand=self.val.press_vcmd,
                                     invalidcommand=self.val.press_ivcmd)

        # temp
        self.temp_lbl = ttk.Label(self.other_lbl_frame,
                                  text='Annual avg local temperature (°C): ')
        self.temp_entry = ttk.Entry(self.other_lbl_frame,
                                    textvariable=self.temp, width=11,
                                    validate='focusout',
                                    validatecommand=self.val.temp_vcmd,
                                    invalidcommand=self.val.temp_ivcmd)

        # output
        self.title_entry = ttk.Entry(self.output_lbl_frame,
                                     textvariable=self.title, width=41)

        self.hz_chart_chkbox = ttk.Checkbutton(self.output_lbl_frame,
                                               text='Plot horizontal sun path',
                                               variable=self.hz_chart,
                                               onvalue=1,
                                               offvalue=0)

        self.vt_chart_chkbox = ttk.Checkbutton(self.output_lbl_frame,
                                               text='Plot vertical sun path',
                                               variable=self.vt_chart,
                                               onvalue=1,
                                               offvalue=0)

        self.plt_point_chkbox = ttk.Checkbutton(self.output_lbl_frame,
                                                text='Plot sun symbol',
                                                variable=self.plt_point,
                                                onvalue=1,
                                                offvalue=0)

        self.csv_data_chkbox = ttk.Checkbutton(self.output_lbl_frame,
                                               text='Save plotting data as csv file',
                                               variable=self.csv_data,
                                               onvalue=1,
                                               offvalue=0)

        self.outputdir_btn = ttk.Button(self.output_lbl_frame,
                                        text='Select output directory',
                                        command=self.__output_directory)
        self.outputdir_lbl = ttk.Label(self.output_lbl_frame,
                                       textvariable=self.outputdir, width=40)

        # buttons
        self.quit_button = ttk.Button(self.button_frame, text='Close',
                                      command=self.main_window.destroy)
        self.getcharts_button = ttk.Button(self.button_frame,
                                           text='Plot Charts',
                                           command=self.__get_charts)

    def __create_frames(self):
        """Create frames"""
        self.button_frame = ttk.Frame(self.main_window)

    def __create_lbl_frames(self):
        """Create label frames"""
        self.date_lbl_frame = ttk.LabelFrame(self.main_window,
                                             text='Date and time',
                                             labelanchor='nw',
                                             padding=5)
        self.location_lbl_frame = ttk.LabelFrame(self.main_window,
                                                 text='Location',
                                                 labelanchor='nw',
                                                 padding=5)
        self.other_lbl_frame = ttk.LabelFrame(self.main_window,
                                              text='Other parameters (not mandatory)',
                                              labelanchor='nw',
                                              padding=5)
        self.output_lbl_frame = ttk.LabelFrame(self.main_window,
                                               text='Output chart options',
                                               labelanchor='nw',
                                               padding=5)

    def __create_layouts(self):
        """Create layouts"""
        # year
        self.year_lbl.grid(column=0, row=1, sticky='E', pady=(0, 10))
        self.year_spin.grid(column=1, row=1, sticky='W', pady=(0, 5))

        # month
        self.month_lbl.grid(column=2, row=1, sticky='E', padx=(10, 1),
                            pady=(0, 5))
        self.month_spin.grid(column=3, row=1, sticky='W', pady=(0, 5))

        # day
        self.day_lbl.grid(column=4, row=1, sticky='E', padx=(10, 1),
                          pady=(0, 5))
        self.day_spin.grid(column=5, row=1, sticky='W', pady=(0, 5))

        # hour
        self.hour_lbl.grid(column=0, row=2, sticky='E')
        self.hour_spin.grid(column=1, row=2, sticky='W')

        # minute
        self.minute_lbl.grid(column=2, row=2, sticky='E', padx=(10, 1))
        self.minute_spin.grid(column=3, row=2, sticky='W')

        # time zone
        self.timezone_lbl.grid(column=4, row=2, sticky='E', padx=(10, 1),
                               pady=(5, 5))
        self.timezone_dropdown.grid(column=5, row=2, sticky='W', pady=(5, 5))

        # daylight
        self.daylight_chkbox.grid(column=1, row=3, sticky='W', columnspan=3)

        # latitude
        self.latitude_lbl.grid(column=0, row=4, sticky='E')
        self.latitude_entry.grid(column=1, row=4, sticky='W')

        # longitude
        self.longitude_lbl.grid(column=2, row=4, sticky='E', padx=(20, 1))
        self.longitude_entry.grid(column=3, row=4, sticky='W')

        # observers elevation
        self.obs_elev_lbl.grid(column=0, row=5, sticky='E', pady=(10, 10))
        self.obs_elev_entry.grid(column=1, row=5, sticky='W')

        # pressure
        self.press_lbl.grid(column=0, row=6, sticky='E', pady=(0, 10))
        self.press_entry.grid(column=1, row=6, sticky='W', pady=(0, 10))

        # temperature
        self.temp_lbl.grid(column=0, row=7, sticky='E', pady=(0, 10))
        self.temp_entry.grid(column=1, row=7, sticky='W', pady=(0, 10))

        # chart title
        self.title_entry.grid(column=0, row=8, sticky='W')

        # horizontal checkbox
        self.hz_chart_chkbox.grid(column=0, row=9, sticky='W', pady=(10, 5))

        # vertical checkbox
        self.vt_chart_chkbox.grid(column=0, row=10, sticky='W', pady=(0, 5))

        # plot sun symbol
        self.plt_point_chkbox.grid(column=0, row=11, sticky='W', pady=(0, 5))

        # data checkbox
        self.csv_data_chkbox.grid(column=0, row=12, sticky='W', pady=(0, 10))

        # output directory
        self.outputdir_btn.grid(column=0, row=13, sticky='W', pady=(0, 10))
        self.outputdir_lbl.grid(column=0, row=14, sticky='W')

        # buttons
        self.quit_button.grid(column=0, row=15, padx=5, pady=(10, 5),
                              sticky='WE')
        self.getcharts_button.grid(column=1, row=15, padx=5, pady=(10, 5),
                                   sticky='WE')

        # label frames
        self.date_lbl_frame.grid(column=0, row=1, padx=5, pady=(10, 10),
                                 sticky='WE')
        self.location_lbl_frame.grid(column=0, row=2, padx=5, pady=(10, 10),
                                     sticky='WE')
        self.other_lbl_frame.grid(column=0, row=3, padx=5, pady=(10, 10),
                                  sticky='WE')
        self.output_lbl_frame.grid(column=0, row=4, padx=5, pady=(10, 10),
                                   sticky='WE')
        # frames
        self.button_frame.grid(column=0, row=5, padx=10, pady=(10, 10),
                               sticky='E')

    def __output_directory(self):
        """Dialog to select output directory"""
        self.outputdir.set(filedialog.askdirectory())

    def __test_input(self, input_string, param_name, is_floating):
        """Test input values before assignment"""
        try:
            if is_floating:
                return float(input_string)
            else:
                return int(input_string)
        except ValueError:
            self.val.error_msgbox(f'Invalid {param_name} value', '')
            return

    def __get_charts(self):
        """Class to create the sun position charts"""
        year = self.__test_input(self.year.get(), 'year', False)
        month = self.__test_input(self.month.get(), 'month', False)
        day = self.__test_input(self.day.get(), 'day', False)
        hour = self.__test_input(self.hour.get(), 'hour', False)
        minute = self.__test_input(self.minute.get(), 'minute', False)
        timezone = int(self.timezone.get())
        daylight = self.daylight.get()
        lat = self.__test_input(self.latitude.get(), 'latitude', True)
        lon = self.__test_input(self.longitude.get(), 'longitude', True)
        obs_elev = 0 if self.obs_elev.get() == '' else \
            self.__test_input(self.obs_elev.get(), 'observer elevation', True)
        press = 0 if self.press.get() == '' else \
            self.__test_input(self.press.get(), 'pressure', True)
        temp = 0 if self.temp.get() == '' else \
            self.__test_input(self.temp.get(), 'temperature', True)
        title = self.title.get()
        hz_chart = self.hz_chart.get()
        vt_chart = self.vt_chart.get()
        plt_point = self.plt_point.get()
        csv_data = self.csv_data.get()
        path = self.outputdir.get()

        # test for date and time correctness
        tmp = [self.val.year_validation(year),
               self.val.month_validation(month),
               self.val.day_validation(day),
               self.val.hour_validation(hour),
               self.val.minute_validation(minute)]

        if not all(tmp):
            self.val.error_msgbox('Invalid date or time value', '')
            return
        else:
            try:
                # seconds are zero
                dt = datetime.datetime(year, month, day, hour, minute, 0)
                if int(daylight) == 1:
                    dt = dt - datetime.timedelta(hours=1)
            except ValueError:
                self.val.error_msgbox('Invalid date or time value', '')
                return

        # test for coordinates correctness
        tmp = [self.val.lat_validation(lat), self.val.lon_validation(lon)]

        if not all(tmp):
            self.val.error_msgbox('Invalid coordinate values', '')
            return

        # test for time zone and latitude congruency
        tz_lat_pairs = ((-12, -180), (-11, -165), (-10, -150), (-9, -135),
                        (-8, -120), (-7, -105), (-6, -90), (-5, -75),
                        (-4, -60), (-3, -45), (-2, -30), (-1, 15), (0, 0),
                        (1, 15), (2, 30), (3, 45), (4, 60), (5, 75),
                        (6, 90), (7, 105), (8, 120), (9, 135), (10, 150),
                        (11, 165), (12, 180))
        tz_width = 4

        for item in tz_lat_pairs:
            if item[0] == timezone:
                low_bound = item[1] - tz_width*15
                upp_bound = item[1] + tz_width*15
                if lon <= low_bound or lon >= upp_bound:
                    txt_main = f'Longitude values must be between {tz_width} hours of the selected time zone'
                    txt_detail = f'For the time zone "UTC {timezone}" the minimum and maximum longitude values are: {low_bound} and {upp_bound}'
                    self.val.error_msgbox(txt_main, txt_detail)
                    return

        sunpos = SunPosition(lat, lon, dt.isoformat(), timezone, obs_elev,
                             press, temp)
        horizon_coords = sunpos.sun_position()
        horizon_coords_point = sunpos.sun_position_point()

        # save azimuth and altitude data as a csv file
        if int(csv_data):
            field_names = ['Dates', 'Azimuths', 'Altitudes']
            file_name = f'SunPositions_{year}-{month}-{day}_{hour}-{minute}.csv'
            sun_positions = []

            for key, value in horizon_coords.items():
                sun_positions.append([key, value[0], value[1]])
            with open(os.path.join(path, file_name), 'w', newline='', encoding='UTF8') as f:
                writer = csv.writer(f)
                writer.writerow(field_names)
                writer.writerows(sun_positions)

        plot = PlotSunPath(horizon_coords, horizon_coords_point, lat,
                           lon, dt, timezone, title, hz_chart, vt_chart,
                           plt_point, path)
        plot.plot_diagrams()


class GUIValidation():
    "GUI validation for the input parameters"
    def __init__(self, main_window):
        self.main_window = main_window

        # year
        self.year_vcmd = (self.main_window.register(self.year_validation),
                          '%P')
        self.year_ivcmd = (self.main_window.register(self.on_year_invalid),)

        # month
        self.month_vcmd = (self.main_window.register(self.month_validation),
                           '%P')
        self.month_ivcmd = (self.main_window.register(self.on_month_invalid),)

        # day
        self.day_vcmd = (self.main_window.register(self.day_validation), '%P')
        self.day_ivcmd = (self.main_window.register(self.on_day_invalid),)

        # hour
        self.hour_vcmd = (self.main_window.register(self.hour_validation),
                          '%P')
        self.hour_ivcmd = (self.main_window.register(self.on_hour_invalid),)

        # minute
        self.minute_vcmd = (
            self.main_window.register(self.minute_validation), '%P')
        self.minute_ivcmd = (
            self.main_window.register(self.on_minute_invalid),)

        # latitude
        self.lat_vcmd = (self.main_window.register(self.lat_validation), '%P')
        self.lat_ivcmd = (self.main_window.register(self.on_lat_invalid),)

        # longitude
        self.lon_vcmd = (self.main_window.register(self.lon_validation), '%P')
        self.lon_ivcmd = (self.main_window.register(self.on_lon_invalid),)

        # observer elevation
        self.obs_elev_vcmd = (
            self.main_window.register(self.obs_elev_validation), '%P')
        self.obs_elev_ivcmd = (
            self.main_window.register(self.on_obs_elev_invalid),)

        # pressure
        self.press_vcmd = (self.main_window.register(self.press_validation),
                           '%P')
        self.press_ivcmd = (self.main_window.register(self.on_press_invalid),)

        # temperature
        self.temp_vcmd = (self.main_window.register(self.temp_validation),
                          '%P')
        self.temp_ivcmd = (self.main_window.register(self.on_temp_invalid),)

    def year_validation(self, year):
        """Validate the year"""
        if int(year) < 1900 or int(year) > 2100:
            return False
        return True

    def on_year_invalid(self):
        """On invalid year"""
        self.error_msgbox('Invalid year',
                          'Year must be in the range 1900 to 2100')

    def month_validation(self, month):
        """Validate the month"""
        if int(month) < 1 or int(month) > 12:
            return False
        return True

    def on_month_invalid(self):
        """On invalid month"""
        self.error_msgbox('Invalid month',
                          'Month must be in the range 1 to 12')

    def day_validation(self, day):
        """Validate the day"""
        if int(day) < 1 or int(day) > 31:
            return False
        return True

    def on_day_invalid(self):
        """On invalid day"""
        self.error_msgbox('Invalid day',
                          'Day must be in the range 1 to 31')

    def hour_validation(self, hour):
        """Validate the hour"""
        if int(hour) < 0 or int(hour) > 23:
            return False
        return True

    def on_hour_invalid(self):
        """On invalid hour"""
        self.error_msgbox('Invalid hour',
                          'Hour must be in the range 0 to 24')

    def on_minute_invalid(self):
        """On invalid minute"""
        self.error_msgbox('Invalid minute',
                          'Minute must be in the range 0 to 59')

    def minute_validation(self, minute):
        """Validate the minute"""
        if int(minute) < 0 or int(minute) > 59:
            return False
        return True

    def lat_validation(self, lat):
        """Validate the latitude"""
        if float(lat) < -90 or float(lat) > 90:
            return False
        return True

    def on_lat_invalid(self):
        """On invalid latitude"""
        self.error_msgbox('Invalid latitude value',
                          'Latitudes must be in the range -90 to 90')

    def lon_validation(self, lon):
        """Validate the longitude"""
        if float(lon) < -180 or float(lon) > 180:
            return False
        return True

    def on_lon_invalid(self):
        """On invalid longitude"""
        self.error_msgbox('Invalid longitude value',
                          'Longitudes must be in the range -180 to 180')

    def obs_elev_validation(self, obs_elev):
        """Validate the observer elevation"""
        if obs_elev == '':
            if float(obs_elev) < -420 or float(obs_elev) > 8850:
                return False
        return True

    def on_obs_elev_invalid(self):
        """On invalid observer elevation"""
        self.error_msgbox('Invalid observer elevation value',
                          'Elevation must be in the range -420 to 8850 meters')

    def press_validation(self, press):
        """Validate the atmospheric pressure"""
        if press == '':
            if float(press) < 850 or float(press) > 1090:
                return False
        return True

    def on_press_invalid(self):
        """On invalid atmospheric pressure"""
        self.error_msgbox('Invalid atmospheric pressure value',
                          'Atmospheric pressure must be in the range 850 to 1090 mbars')

    def temp_validation(self, temp):
        """Validate the temperature"""
        if temp == '':
            if float(temp) < -90 or float(temp) > 57:
                return False
            return True

    def on_temp_invalid(self):
        """On invalid temperature"""
        self.error_msgbox('Invalid temperature value',
                          'Atmospheric pressure must be in the range -90 to 57 C')

    def error_msgbox(self, message, detail):
        """Error message box"""
        messagebox.showerror(title='Input error', message=message,
                             detail=detail)


if __name__ == '__main__':
    app = SunPathGUI()
    app.main_window.mainloop()


"""This module contains a single class used to plot the vertical and horizontal
sun-path diagrams"""

import os
from math import radians, degrees
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.ticker import AutoMinorLocator
import sun_path_diagrams.config as cfg


class PlotSunPath:
    """This class will plot the horizontal and vertical sun-path diagrams"""
    def __init__(self, horizon_coords, horizon_coords_point, lat, lon, dt,
                 timezone, title, hz_chart, vt_chart, plt_point, path):
        self.horizon_coords = horizon_coords
        self.horizon_coords_point = horizon_coords_point
        self.lat = lat
        self.lon = lon
        self.date = dt.strftime('%Y-%b-%d')
        self.time = dt.strftime('%H:%M')
        self.timezone = timezone
        self.title = title
        self.hz_chart = int(hz_chart)
        self.vt_chart = int(vt_chart)
        self.plt_point = int(plt_point)
        self.path = path

    def plot_diagrams(self):
        """Plots the vertical and horizontal sun path diagrams"""
        chart_data = self.__get_chart_data()
        if self.hz_chart:
            self.__horizontal_sunpath(chart_data)
        if self.vt_chart:
            self.__vertical_sunpath(chart_data)

    def __vertical_sunpath(self, chart_data):
        """Plots the vertical sun-path diagram"""
        fig, ax = plt.subplots()
        fig.set_size_inches(cfg.fig_zize['width'], cfg.fig_zize['height'])

        # create lines for legend
        lines = [matplotlib.lines.Line2D([0], [0], linestyle=item[0],
                 color=cfg.equisols_symb['color']) for item in
                 cfg.equisols.values()]

        # create labels for legend
        labels = [item[1] for item in cfg.equisols.values()]

        for key, value in chart_data.items():
            if key[0] == 'p':
                azim = value[0]
                alt = value[1]
                # plot user curve
                ax.plot([degrees(i) for i in azim], alt,
                        color=cfg.user_symb['color'],
                        linewidth=cfg.user_symb['linewidth'])
                # line and label for legend
                lines.append(matplotlib.lines.Line2D([0], [0],
                             color=cfg.user_symb['color']))
                labels.append(key[1:4] + ' ' + key[4:6])

                # plot user's mark
                if self.plt_point:
                    azim = self.horizon_coords_point[0]
                    alt = self.horizon_coords_point[1]
                    ax.plot(azim, alt, color=cfg.user_symb_point['color'],
                            marker=cfg.user_symb_point['marker'],
                            markersize=cfg.user_symb_point['markersize'])
                    lines.append(matplotlib.lines.Line2D([0], [0],
                                 color=cfg.user_symb_point['color'],
                                 marker=cfg.user_symb_point['marker'],
                                 markersize=cfg.user_symb_point['markersize'],
                                 linewidth=cfg.user_symb_point['linewidth']))
                    labels.append('Sun position')
            elif key == 'hours':
                # plot analemmas
                analemmas = self.__get_analemmas(value)
                for _, value_ in analemmas.items():
                    ax.plot([degrees(i) for i in value_[0]], value_[1],
                            color=cfg.ana_symb['color'],
                            linestyle=cfg.ana_symb['linestyle'],
                            linewidth=cfg.ana_symb['linewidth'])
                # plot hour labels
                for item in value:
                    if item[0] == 'Jun21':
                        # range starts at 1 so we don't plot the 25th hour
                        for j in range(1, len(item[1])):
                            ax.annotate(item[3][j],
                                        xy=(degrees(item[1][j]), item[2][j]),
                                        xytext=cfg.hour_symb['xytext'],
                                        textcoords='offset points',
                                        color=cfg.hour_symb['color'],
                                        size=cfg.hour_symb['size'])
            else:
                azim = value[0]
                alt = value[1]
                # plot equinoxes and solstices curves
                ax.plot([degrees(i) for i in azim], alt,
                        color=cfg.equisols_symb['color'],
                        linewidth=cfg.equisols_symb['linewidth'],
                        linestyle=cfg.equisols[key][0])

        ax.xaxis.set_minor_locator(AutoMinorLocator(2))
        ax.yaxis.set_minor_locator(AutoMinorLocator(2))
        ax.tick_params(axis='both', labelsize='8')
        ax.tick_params(which='minor', length=2)

        ax.set_aspect(3)

        ax.set_ylim(bottom=0, top=90)
        y_ticks, y_labels = self.__get_yticks(0, 91, 10)
        ax.set_yticks(y_ticks)
        ax.set_yticklabels(y_labels)

        ax.set_xlim(left=0, right=radians(360))
        x_ticks, x_labels = self.__get_xticks(0, 361, 30)
        ax.set_xticks(x_ticks)
        ax.set_xticklabels(x_labels)

        ax.grid(color=cfg.grid_symb['color1'],
                linestyle=cfg.grid_symb['linestyle'],
                linewidth=cfg.grid_symb['linewidth'])

        ax.set_title(self.title)
        ax.set_xlabel('Azimuth')
        ax.set_ylabel('Altitude')

        fig.legend(lines, labels, loc='lower right', fontsize='small',
                   handlelength=4, frameon=False)
        text = f'Vertical Sun Path Diagram \nLatitude: {self.lat}° \
                \nLongitude: {self.lon}° \nDate: {self.date}'
        if self.plt_point:
            text = text + '\nTime: ' + self.time
        fig.text(0.01, 0.02, text, fontsize='small', linespacing=1.5)

        file_name = f'VerticalSunPath_{self.date}.png'
        plt.savefig(os.path.join(self.path, file_name))
        #plt.show()
        plt.close()

    def __horizontal_sunpath(self, chart_data):
        """Plots the horizontal sun-path diagram"""
        fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
        fig.set_size_inches(cfg.fig_zize['width'], cfg.fig_zize['height'])

        # create lines for legend
        lines = [matplotlib.lines.Line2D([0], [0], linestyle=item[0],
                 color=cfg.equisols_symb['color']) for item in
                 cfg.equisols.values()]

        # create labels for legend
        labels = [item[1] for item in cfg.equisols.values()]

        for key, value in chart_data.items():
            if key[0] == 'p':
                # calculate the revised azimuth and altitude
                azim, alt = self.__revised_azim_alt(value[0], value[1],
                                                    self.lat)
                # plot date user curve
                ax.plot(azim, alt, color=cfg.user_symb['color'],
                        linewidth=cfg.user_symb['linewidth'])

                # plot hours in the date user curve
                # ax.plot(azim, alt, color=cfg.user_symb['color'],
                #         marker=cfg.user_symb['marker'])

                # line and label for legend
                lines.append(matplotlib.lines.Line2D([0], [0],
                             color=cfg.user_symb['color']))
                labels.append(key[1:4] + ' ' + key[4:6])

                # plot user's mark
                if self.plt_point:
                    azim = radians(self.horizon_coords_point[0])
                    alt = self.horizon_coords_point[1]
                    ax.plot(azim, alt, color=cfg.user_symb_point['color'],
                            marker=cfg.user_symb_point['marker'],
                            markersize=cfg.user_symb_point['markersize'])
                    lines.append(matplotlib.lines.Line2D([0], [0],
                                 color=cfg.user_symb_point['color'],
                                 marker=cfg.user_symb_point['marker'],
                                 markersize=cfg.user_symb_point['markersize'],
                                 linewidth=cfg.user_symb_point['linewidth']))
                    labels.append('Sun position')

                # draw minor ticks every 5 degrees
                az, al = self.__get_hz_minor_ticks(5)
                ax.plot(az, al, color=cfg.grid_symb['color2'],
                        linewidth=cfg.grid_symb['linewidth'])
            elif key == 'hours':
                # plot analemmas
                analemmas = self.__get_analemmas(value)
                for _, value_ in analemmas.items():
                    ax.plot(value_[0], value_[1], color=cfg.ana_symb['color'],
                            linestyle=cfg.ana_symb['linestyle'],
                            linewidth=cfg.ana_symb['linewidth'])
                # plot hour labels
                for item in value:
                    # range starts at 1 so we don't plot the 25th hour
                    for j in range(1, len(item[1])):
                        if item[0] == 'Jun21' and self.lat >= 0:
                            ax.annotate(item[3][j],
                                        xy=(item[1][j], item[2][j]),
                                        xytext=self.__repo_labels(item[1][j],
                                                                  self.lat),
                                        textcoords='offset points',
                                        color=cfg.hour_symb['color'],
                                        size=cfg.hour_symb['size'],
                                        horizontalalignment='center',
                                        verticalalignment='center')
                        if item[0] == 'Dec21' and self.lat < 0:
                            ax.annotate(item[3][j],
                                        xy=(item[1][j], item[2][j]),
                                        xytext=self.__repo_labels(item[1][j],
                                                                  self.lat),
                                        textcoords='offset points',
                                        color=cfg.hour_symb['color'],
                                        size=cfg.hour_symb['size'],
                                        horizontalalignment='center',
                                        verticalalignment='center')
            else:
                azim, alt = self.__revised_azim_alt(value[0], value[1],
                                                    self.lat)
                ax.plot(azim, alt, color=cfg.equisols_symb['color'],
                        linewidth=cfg.equisols_symb['linewidth'],
                        linestyle=cfg.equisols[key][0])

        ax.set_title(self.title)
        ax.set_theta_zero_location('N')
        ax.set_theta_direction(-1)
        ax.set(aspect='equal', rorigin=90)
        ax.set_rmax(0)
        ax.set_rmin(90)
        r_ticks, r_labels = self.__get_yticks(10, 81, 10)
        ax.set_rgrids(r_ticks, r_labels, angle=cfg.grid_symb['angle'],
                      color=cfg.grid_symb['color3'],
                      size=cfg.grid_symb['size'])

        theta_ticks, theta_labels = self.__get_xticks(0, 360, 15)
        ax.set_thetagrids(theta_ticks, theta_labels,
                          color=cfg.grid_symb['color2'],
                          size=cfg.grid_symb['size'])

        ax.grid(color=cfg.grid_symb['color1'],
                linestyle=cfg.grid_symb['linestyle'],
                linewidth=cfg.grid_symb['linewidth'])

        fig.legend(lines, labels, loc='lower right', fontsize='small',
                   handlelength=4, frameon=False)
        text = f'Horizontal Sun Path Diagram \nLatitude: {self.lat}° \
                \nLongitude: {self.lon}° \nDate: {self.date}'
        if self.plt_point:
            text = text + '\nTime: ' + self.time
        fig.text(0.01, 0.02, text, fontsize='small', linespacing=1.5)

        file_name = f'HorizontalSunPath_{self.date}.png'
        plt.savefig(os.path.join(self.path, file_name))
        #plt.show()
        plt.close()

    def __get_chart_data(self):
        """Process the data for the plotting functions"""
        whole_hours = []
        key_dates = ['Jun21', 'Mar21', 'Dec21']
        chart_data = {}

        for key, value in self.horizon_coords.items():
            azimuths = value[0]
            altitudes = value[1]

            if key in key_dates:
                # convert azimuth degrees to radians
                azimuth_rad = [radians(i) for i in azimuths]
                chart_data[key] = (azimuth_rad, altitudes)

            # plot input's date
            if key[0] == 'p':
                # convert azimuth degrees to radians
                azimuth_rad = [radians(i) for i in azimuths]
                chart_data[key] = (azimuth_rad, altitudes)

            off_ut = self.timezone

            # get the coordinates for whole hours
            azims, alts, hours = self.__get_hours(azimuths, altitudes, off_ut)
            whole_hours.append((key, azims, alts, hours))

        chart_data['hours'] = whole_hours

        return chart_data

    def __revised_azim_alt(self, azim, alt, lat):
        """For latitudes between -30 and 30 degrees there are lines that plots
        across the circle. This method adds new coordinates to move these lines
        off the circle plot"""

        if lat > -30 and lat < 30:
            alt_sort = alt.copy()
            alt_sort.sort()
            for item in alt_sort:
                ndx = alt.index(item)
                if degrees(azim[ndx]) >= 270:
                    azim.insert(ndx + 1, 5.5)
                    alt.insert(ndx + 1, -89)
                    azim.insert(ndx + 2, 6.27)
                    alt.insert(ndx + 2, -89)
                    return azim, alt
                elif degrees(azim[ndx]) >= 180 and degrees(azim[ndx]) < 270:
                    azim.insert(ndx + 1, 3.14)
                    alt.insert(ndx + 1, -89)
                    azim.insert(ndx + 2, 3.67)
                    alt.insert(ndx + 2, -89)
                    return azim, alt

        return azim, alt

    def __get_hours(self, azimuths, altitudes, off_ut):
        """Calculate whole hours positions"""
        azims = []
        alts = []
        hours = []

        count = off_ut - 1
        for i in azimuths:
            index = azimuths.index(i)
            if count == 24:
                count = 0
            count = count + 1
            azims.append(radians(i))
            alts.append(altitudes[index])
            hours.append(count)

        hour_pairs = [(-1, 23), (-2, 22), (-3, 21), (-4, 20), (-5, 19),
                      (-6, 18), (-7, 17), (-8, 16), (-9, 15), (-10, 14),
                      (-11, 13), (-12, 12)]

        for item in hour_pairs:
            if item[0] in hours:
                ndx = hours.index(item[0])
                hours[ndx] = item[1]

        return azims, alts, hours

    def __get_analemmas(self, hours):
        """Get the analemma coordinates"""
        min_hour = 24
        max_hour = 0

        for i in hours:
            if min(i[3]) < min_hour:
                min_hour = min(i[3])
            if max(i[3]) > max_hour:
                max_hour = max(i[3])

        analemmas = {}

        for i in range(min_hour, max_hour + 1):
            azims = []
            alts = []
            for j in hours:
                if i in j[3]:
                    p = j[3].index(i)
                    azims.append(j[1][p])
                    alts.append(j[2][p])
            analemmas[str(i)] = (azims, alts)

        return analemmas

    def __get_hz_minor_ticks(self, interval):
        """Given the interval, gets the minor ticks for the horizontal sun-path 
        diagram. Note that minor thicks are not available for polar charts,
        hence this hack"""
        ticks = range(0, 360, interval)
        tmp = [radians(i) for i in ticks]

        return [tmp, tmp], [[1]*int(360/interval), [0]*int(360/interval)]

    def __get_xticks(self, start, end, interval):
        """Given the start, end and interval, get the ticks and labels for the
        x-axis and theta-axis"""
        ticks = range(start, end, interval)
        labels = []
        for i in ticks:
            if i == 0 or i == 360:
                labels.append('N')
            elif i == 90:
                labels.append('E')
            elif i == 180:
                labels.append('S')
            elif i == 270:
                labels.append('W')
            else:
                labels.append(str(i) + '°')

        return ticks, labels

    def __get_yticks(self, start, end, interval):
        """Given the start, end and interval, get the ticks and labels for the 
        y-axis and r-axis"""
        ticks = range(start, end, interval)
        labels = [str(i) + '°' for i in ticks]

        return ticks, labels

    def __repo_labels(self, x, lat):
        """Reposition the hour labels"""
        if lat >= 60 or lat <= -60:
            x = degrees(x)
            if x >= 345:
                pos = (0, -6)
            elif x < 15:
                pos = (0, -6)
            elif x >= 15 and x < 45:
                pos = (-2, -6)
            elif x >= 45 and x < 75:
                pos = (-4, -6)
            elif x >= 75 and x < 105:
                pos = (-5, 0)
            elif x >= 105 and x < 135:
                pos = (-4, 6)
            elif x >= 135 and x < 165:
                pos = (-2, 6)
            elif x >= 165 and x < 195:
                pos = (0, 6)
            elif x >= 195 and x < 225:
                pos = (2, 6)
            elif x >= 225 and x < 255:
                pos = (4, 6)
            elif x >= 255 and x < 285:
                pos = (5, 0)
            elif x >= 285 and x < 315:
                pos = (4, -6)
            elif x >= 315 and x < 345:
                pos = (2, -6)
            else:
                pos = (0, 6)
        elif lat >= 0 and lat < 60:
            pos = (0, 6)
        elif lat < 0 and lat > -60:
            pos = (0, -6)
        else:
            pos = (0, 6)

        return pos